@@ -18,16 +18,33 @@ import signal
 import subprocess
 
 from bzrlib import (
-        branch,
-        bzrdir,
-        errors,
-        merge,
-        revisionspec,
-        tag,
-        transport,
-        ui,
-        urlutils,
-        )
+    branch,
+    bzrdir,
+    errors,
+    merge,
+    revision,
+    revisionspec,
+    tag,
+    trace,
+    transport,
+    urlutils,
+    )
+
+try:
+    from bzrlib.errors import NoWhoami
+except ImportError:
+    NoWhoami = object()
+
+
+try:
+    from debian import changelog
+except ImportError:
+    from debian_bundle import changelog
+
+try:
+    MergeIntoMerger = merge.MergeIntoMerger
+except (AttributeError, NameError):
+    from bzrlib.plugins.builder.bzrlibbackports import MergeIntoMerger
 
 
 def subprocess_setup():
@@ -35,15 +52,35 @@ def subprocess_setup():
 
 
 MERGE_INSTRUCTION = "merge"
+NEST_PART_INSTRUCTION = "nest-part"
 NEST_INSTRUCTION = "nest"
 RUN_INSTRUCTION = "run"
 
+SAFE_INSTRUCTIONS = [
+    MERGE_INSTRUCTION, NEST_PART_INSTRUCTION, NEST_INSTRUCTION]
+
 TIME_VAR = "{time}"
+DATE_VAR = "{date}"
 REVNO_VAR = "{revno}"
 REVNO_PARAM_VAR = "{revno:%s}"
 DEBUPSTREAM_VAR = "{debupstream}"
 
 ok_to_preserve = [DEBUPSTREAM_VAR]
+# The variables that don't require substitution in their name
+simple_vars = [TIME_VAR, DATE_VAR, REVNO_VAR, DEBUPSTREAM_VAR]
+
+
+def check_expanded_deb_version(base_branch):
+    checked_version = base_branch.deb_version
+    for token in ok_to_preserve:
+        checked_version = checked_version.replace(token, "")
+    if "{" in checked_version:
+        available_tokens = simple_vars
+        for name in base_branch.list_branch_names():
+            available_tokens.append(REVNO_PARAM_VAR % name)
+        raise errors.BzrCommandError("deb-version not fully "
+                "expanded: %s. Valid substitutions are: %s"
+                % (base_branch.deb_version, available_tokens))
 
 
 class CommandFailedError(errors.BzrError):
@@ -153,7 +190,7 @@ def pull_or_branch(tree_to, br_to, br_from, to_transport, revision_id,
     return tree_to, br_to
 
 
-def merge_branch(child_branch, tree_to, br_to):
+def merge_branch(child_branch, tree_to, br_to, possible_transports=None):
     """Merge the branch specified by child_branch.
 
     :param child_branch: the RecipeBranch to retrieve the branch and revision to
@@ -161,82 +198,138 @@ def merge_branch(child_branch, tree_to, br_to):
     :param tree_to: the WorkingTree to merge in to.
     :param br_to: the Branch to merge in to.
     """
+    if child_branch.branch is None:
+        child_branch.branch = branch.Branch.open(child_branch.url,
+                possible_transports=possible_transports)
+    child_branch.branch.lock_read()
+    try:
+        tag._merge_tags_if_possible(child_branch.branch, br_to)
+        if child_branch.revspec is not None:
+            merge_revspec = revisionspec.RevisionSpec.from_string(
+                    child_branch.revspec)
+            try:
+                merge_revid = merge_revspec.as_revision_id(
+                    child_branch.branch)
+            except errors.InvalidRevisionSpec, e:
+                # Give the user a hint if they didn't mean to speciy
+                # a revspec.
+                e.extra = (". Did you not mean to specify a revspec "
+                    "at the end of the merge line?")
+                raise e
+        else:
+            merge_revid = child_branch.branch.last_revision()
+        child_branch.revid = merge_revid
+        merger = merge.Merger.from_revision_ids(None, tree_to, merge_revid,
+                other_branch=child_branch.branch, tree_branch=br_to)
+        merger.merge_type = merge.Merge3Merger
+        if (merger.base_rev_id == merger.other_rev_id and
+                merger.other_rev_id is not None):
+            # Nothing to do.
+            return
+        conflict_count = merger.do_merge()
+        merger.set_pending()
+        if conflict_count:
+            # FIXME: better reporting
+            raise errors.BzrCommandError("Conflicts from merge")
+        config = br_to.get_config()
+        try:
+            committer = config.username()
+        except NoWhoami:
+            committer = "bzr-builder <nobody@example.com>"
+        tree_to.commit("Merge %s" %
+                urlutils.unescape_for_display(child_branch.url, 'utf-8'),
+                committer=committer)
+    finally:
+        child_branch.branch.unlock()
+
+
+def nest_part_branch(child_branch, tree_to, br_to, subpath, target_subdir=None):
+    """Merge the branch subdirectory specified by child_branch.
+
+    :param child_branch: the RecipeBranch to retrieve the branch and revision to
+            merge from.
+    :param tree_to: the WorkingTree to merge in to.
+    :param br_to: the Branch to merge in to.
+    :param subpath: only merge files from branch that are from this path.
+        e.g. subpath='/debian' will only merge changes from that directory.
+    :param target_subdir: (optional) directory in target to merge that
+        subpath into.  Defaults to basename of subpath.
+    """
     merge_from = branch.Branch.open(child_branch.url)
     merge_from.lock_read()
     try:
-        pb = ui.ui_factory.nested_progress_bar()
+        if child_branch.revspec is not None:
+            merge_revspec = revisionspec.RevisionSpec.from_string(
+                    child_branch.revspec)
+            merge_revid = merge_revspec.as_revision_id(merge_from)
+        else:
+            merge_revid = merge_from.last_revision()
+        child_branch.revid = merge_revid
+        other_tree = merge_from.basis_tree()
+        other_tree.lock_read()
         try:
-            tag._merge_tags_if_possible(merge_from, br_to)
-            if child_branch.revspec is not None:
-                merge_revspec = revisionspec.RevisionSpec.from_string(
-                        child_branch.revspec)
-                merge_revid = merge_revspec.as_revision_id(merge_from)
-            else:
-                merge_revid = merge_from.last_revision()
-            child_branch.revid = merge_revid
-            merger = merge.Merger.from_revision_ids(pb, tree_to, merge_revid,
-                    other_branch=merge_from, tree_branch=br_to)
-            merger.merge_type = merge.Merge3Merger
-            if (merger.base_rev_id == merger.other_rev_id and
-                    merger.other_rev_id is not None):
-                # Nothing to do.
-                return
+            if target_subdir is None:
+                target_subdir = os.path.basename(subpath)
+            merger = MergeIntoMerger(this_tree=tree_to, other_tree=other_tree,
+                other_branch=merge_from, target_subdir=target_subdir,
+                source_subpath=subpath, other_rev_id=merge_revid)
+            merger.set_base_revision(revision.NULL_REVISION, merge_from)
             conflict_count = merger.do_merge()
             merger.set_pending()
-            if conflict_count:
-                # FIXME: better reporting
-                raise errors.BzrCommandError("Conflicts from merge")
-            tree_to.commit("Merge %s" %
-                    urlutils.unescape_for_display(
-                        child_branch.url, 'utf-8'))
         finally:
-            pb.finished()
+            other_tree.unlock()
     finally:
         merge_from.unlock()
 
+    if conflict_count:
+        # FIXME: better reporting
+        raise errors.BzrCommandError("Conflicts from merge")
+    tree_to.commit("Merge %s of %s" %
+        (subpath, urlutils.unescape_for_display(child_branch.url, 'utf-8')))
 
-def update_branch(base_branch, tree_to, br_to, to_transport):
-    from_location = base_branch.url
-    accelerator_tree, br_from = bzrdir.BzrDir.open_tree_or_branch(
-                        from_location)
-    br_from.lock_read()
+
+def update_branch(base_branch, tree_to, br_to, to_transport,
+        possible_transports=None):
+    if base_branch.branch is None:
+        base_branch.branch = branch.Branch.open(base_branch.url,
+                possible_transports=possible_transports)
+    base_branch.branch.lock_read()
     try:
         if base_branch.revspec is not None:
             revspec = revisionspec.RevisionSpec.from_string(
                     base_branch.revspec)
-            revision_id = revspec.as_revision_id(br_from)
+            revision_id = revspec.as_revision_id(base_branch.branch)
         else:
-            revision_id = br_from.last_revision()
+            revision_id = base_branch.branch.last_revision()
         base_branch.revid = revision_id
-        tree_to, br_to = pull_or_branch(tree_to, br_to, br_from,
+        tree_to, br_to = pull_or_branch(tree_to, br_to, base_branch.branch,
                 to_transport, revision_id,
-                accelerator_tree=accelerator_tree,
-                possible_transports=[to_transport])
+                possible_transports=possible_transports)
     finally:
-        br_from.unlock()
+        base_branch.branch.unlock()
     return tree_to, br_to
 
 
 def _resolve_revisions_recurse(new_branch, substitute_revno,
         if_changed_from=None):
     changed = False
-    br_from = branch.Branch.open(new_branch.url)
-    br_from.lock_read()
+    new_branch.branch = branch.Branch.open(new_branch.url)
+    new_branch.branch.lock_read()
     try:
         if new_branch.revspec is not None:
             revspec = revisionspec.RevisionSpec.from_string(
                     new_branch.revspec)
-            revision_id = revspec.as_revision_id(br_from)
+            revision_id = revspec.as_revision_id(new_branch.branch)
         else:
-            revision_id = br_from.last_revision()
+            revision_id = new_branch.branch.last_revision()
         new_branch.revid = revision_id
         def get_revno():
             try:
-                revno = br_from.revision_id_to_revno(revision_id)
+                revno = new_branch.branch.revision_id_to_revno(revision_id)
                 return str(revno)
             except errors.NoSuchRevision:
                 # We need to load and use the full revno map after all
-                result = br_from.get_revision_id_to_revno_map().get(
+                result = new_branch.branch.get_revision_id_to_revno_map().get(
                         revision_id)
             if result is None:
                 return result
@@ -249,9 +342,9 @@ def _resolve_revisions_recurse(new_branch, substitute_revno,
                 changed_revspec = revisionspec.RevisionSpec.from_string(
                         if_changed_from.revspec)
                 changed_revision_id = changed_revspec.as_revision_id(
-                        br_from)
+                        new_branch.branch)
             else:
-                changed_revision_id = br_from.last_revision()
+                changed_revision_id = new_branch.branch.last_revision()
             if revision_id != changed_revision_id:
                 changed = True
         for index, instruction in enumerate(new_branch.child_branches):
@@ -267,7 +360,7 @@ def _resolve_revisions_recurse(new_branch, substitute_revno,
                     changed = child_changed
         return changed
     finally:
-        br_from.unlock()
+        new_branch.branch.unlock()
 
 
 def resolve_revisions(base_branch, if_changed_from=None):
@@ -295,29 +388,20 @@ def resolve_revisions(base_branch, if_changed_from=None):
             if_changed_from=if_changed_from_revisions)
     if not changed:
         changed = changed_revisions
-    checked_version = base_branch.deb_version
-    for token in ok_to_preserve:
-        checked_version = checked_version.replace(token, "")
-    if "{" in checked_version:
-        raise errors.BzrCommandError("deb-version not fully "
-                "expanded: %s" % base_branch.deb_version)
+    check_expanded_deb_version(base_branch)
     if if_changed_from is not None and not changed:
         return False
     return True
 
 
-def build_tree(base_branch, target_path):
-    """Build the RecipeBranch at a path.
-
-    Follow the instructions embodied in RecipeBranch and build a tree
-    based on them rooted at target_path. If target_path exists and
-    is the root of the branch then the branch will be updated based on
-    what the RecipeBranch requires.
-
-    :param base_branch: a RecipeBranch to build.
-    :param target_path: the path to the base of the desired output.
-    """
-    to_transport = transport.get_transport(target_path)
+def _build_inner_tree(base_branch, target_path, possible_transports=None):
+    revision_of = ""
+    if base_branch.revspec is not None:
+        revision_of = "revision '%s' of " % base_branch.revspec
+    trace.note("Retrieving %s'%s' to put at '%s'."
+        % (revision_of, base_branch.url, target_path))
+    to_transport = transport.get_transport(target_path,
+            possible_transports=possible_transports)
     try:
         tree_to, br_to = bzrdir.BzrDir.open_tree_or_branch(target_path)
         # Should we commit any changes in the tree here? If we don't
@@ -332,7 +416,7 @@ def build_tree(base_branch, target_path):
             br_to.lock_write()
         try:
             tree_to, br_to = update_branch(base_branch, tree_to, br_to,
-                    to_transport)
+                    to_transport, possible_transports=possible_transports)
             for instruction in base_branch.child_branches:
                 instruction.apply(target_path, tree_to, br_to)
         finally:
@@ -344,88 +428,122 @@ def build_tree(base_branch, target_path):
             tree_to.unlock()
 
 
-def _add_child_branches_to_manifest(child_branches, indent_level):
-    manifest = ""
-    for instruction in child_branches:
-        child_branch = instruction.recipe_branch
-        nest_location = instruction.nest_path
-        if child_branch is None:
-            manifest += "%s%s %s\n" % ("  " * indent_level, RUN_INSTRUCTION,
-                    nest_location)
-        else:
-            assert child_branch.revid is not None, "Branch hasn't been built"
-            if nest_location is not None:
-                manifest += "%s%s %s %s %s revid:%s\n" % \
-                             ("  " * indent_level, NEST_INSTRUCTION,
-                              child_branch.name,
-                              child_branch.url, nest_location,
-                              child_branch.revid)
-                manifest += _add_child_branches_to_manifest(
-                        child_branch.child_branches, indent_level+1)
-            else:
-                manifest += "%s%s %s %s revid:%s\n" % \
-                             ("  " * indent_level, MERGE_INSTRUCTION,
-                              child_branch.name,
-                              child_branch.url, child_branch.revid)
-    return manifest
+def build_tree(base_branch, target_path, possible_transports=None):
+    """Build the RecipeBranch at a path.
 
+    Follow the instructions embodied in RecipeBranch and build a tree
+    based on them rooted at target_path. If target_path exists and
+    is the root of the branch then the branch will be updated based on
+    what the RecipeBranch requires.
 
-def build_manifest(base_branch):
-    manifest = "# bzr-builder format %s deb-version " % str(base_branch.format)
-    # TODO: should we store the expanded version that was used?
-    manifest += "%s\n" % (base_branch.deb_version,)
-    assert base_branch.revid is not None, "Branch hasn't been built"
-    manifest += "%s revid:%s\n" % (base_branch.url, base_branch.revid)
-    manifest += _add_child_branches_to_manifest(base_branch.child_branches, 0)
-    # Sanity check.
-    # TODO: write a function that compares the result of this parse with
-    # the branch that we built it from.
-    RecipeParser(manifest).parse()
-    return manifest
+    :param base_branch: a RecipeBranch to build.
+    :param target_path: the path to the base of the desired output.
+    """
+    trace.note("Building tree.")
+    _build_inner_tree(base_branch, target_path,
+            possible_transports=possible_transports)
 
 
 class ChildBranch(object):
     """A child branch in a recipe.
-    
+
     If the nest path is not None it is the path relative to the recipe branch
     where the child branch should be placed.  If it is None then the child
     branch should be merged instead of nested.
     """
 
+    can_have_children = False
+
     def __init__(self, recipe_branch, nest_path=None):
         self.recipe_branch = recipe_branch
         self.nest_path = nest_path
-        
-    def apply(self, target_path, tree_to, br_to):
+
+    def apply(self, target_path, tree_to, br_to, possible_transports=None):
         raise NotImplementedError(self.apply)
 
     def as_tuple(self):
         return (self.recipe_branch, self.nest_path)
 
+    def _get_revid_part(self):
+        if self.recipe_branch.revid is not None:
+            revid_part = " revid:%s" % self.recipe_branch.revid
+        elif self.recipe_branch.revspec is not None:
+            revid_part = " %s" % self.recipe_branch.revspec
+        else:
+            revid_part = ""
+        return revid_part
+
 
 class CommandInstruction(ChildBranch):
 
-    def apply(self, target_path, tree_to, br_to):
+    def apply(self, target_path, tree_to, br_to, possible_transports=None):
         # it's a command
+        trace.note("Running '%s' in '%s'." % (self.nest_path, target_path))
         proc = subprocess.Popen(self.nest_path, cwd=target_path,
             preexec_fn=subprocess_setup, shell=True, stdin=subprocess.PIPE)
         proc.communicate()
         if proc.returncode != 0:
             raise CommandFailedError(self.nest_path)
 
+    def as_text(self):
+        return "%s %s" % (RUN_INSTRUCTION, self.nest_path)
+
 
 class MergeInstruction(ChildBranch):
 
+    def apply(self, target_path, tree_to, br_to, possible_transports=None):
+        revision_of = ""
+        if self.recipe_branch.revspec is not None:
+            revision_of = "revision '%s' of " % self.recipe_branch.revspec
+        trace.note("Merging %s'%s' in to '%s'."
+                % (revision_of, self.recipe_branch.url, target_path))
+        merge_branch(self.recipe_branch, tree_to, br_to,
+                possible_transports=possible_transports)
+
+    def as_text(self):
+        revid_part = self._get_revid_part()
+        return "%s %s %s%s" % (
+            MERGE_INSTRUCTION, self.recipe_branch.name,
+            self.recipe_branch.url, revid_part)
+
+
+class NestPartInstruction(ChildBranch):
+
+    def __init__(self, recipe_branch, subpath, target_subdir):
+        ChildBranch.__init__(self, recipe_branch)
+        self.subpath = subpath
+        self.target_subdir = target_subdir
+
     def apply(self, target_path, tree_to, br_to):
-        merge_branch(self.recipe_branch, tree_to, br_to)
+        nest_part_branch(self.recipe_branch, tree_to, br_to, self.subpath,
+            self.target_subdir)
+
+    def as_text(self):
+        revid_part = self._get_revid_part()
+        if self.target_subdir is not None:
+            target_revid_part = " %s%s" % (
+                self.target_subdir, revid_part)
+        else:
+            target_revid_part = revid_part
+        return "%s %s %s %s%s" % (
+            NEST_PART_INSTRUCTION, self.recipe_branch.name,
+            self.recipe_branch.url, self.subpath, target_revid_part)
 
 
 class NestInstruction(ChildBranch):
 
-    def apply(self, target_path, tree_to, br_to):
-        # FIXME: pass possible_transports around
-        build_tree(self.recipe_branch,
-            target_path=os.path.join(target_path, self.nest_path))
+    can_have_children = True
+
+    def apply(self, target_path, tree_to, br_to, possible_transports=None):
+        _build_inner_tree(self.recipe_branch,
+            target_path=os.path.join(target_path, self.nest_path),
+            possible_transports=possible_transports)
+
+    def as_text(self):
+        revid_part = self._get_revid_part()
+        return "%s %s %s %s%s" % (
+            NEST_INSTRUCTION, self.recipe_branch.name,
+            self.recipe_branch.url, self.nest_path, revid_part)
 
 
 class RecipeBranch(object):
@@ -435,9 +553,12 @@ class RecipeBranch(object):
     root branch), and optionally child branches that are either merged
     or nested.
 
-    The child_branches attribute is a list of tuples of ChildBranch objects. 
+    The child_branches attribute is a list of tuples of ChildBranch objects.
     The revid attribute records the revid that the url and revspec resolved
     to when the RecipeBranch was built, or None if it has not been built.
+
+    :ivar revid: after this recipe branch has been built this is set to the
+        revision ID that was merged/nested from the branch at self.url.
     """
 
     def __init__(self, name, url, revspec=None):
@@ -453,6 +574,7 @@ class RecipeBranch(object):
         self.revspec = revspec
         self.child_branches = []
         self.revid = None
+        self.branch = None
 
     def merge_branch(self, branch):
         """Merge a child branch in to this one.
@@ -461,6 +583,18 @@ class RecipeBranch(object):
         """
         self.child_branches.append(MergeInstruction(branch))
 
+    def nest_part_branch(self, branch, subpath=None, target_subdir=None):
+        """Merge subdir of a child branch into this one.
+
+        :param branch: the RecipeBranch to merge.
+        :param subpath: only merge files from branch that are from this path.
+            e.g. subpath='/debian' will only merge changes from that directory.
+        :param target_subdir: (optional) directory in target to merge that
+            subpath into.  Defaults to basename of subpath.
+        """
+        self.child_branches.append(
+            NestPartInstruction(branch, subpath, target_subdir))
+
     def nest_branch(self, location, branch):
         """Nest a child branch in to this one.
 
@@ -497,10 +631,32 @@ class RecipeBranch(object):
             if ((child_branch is None and other_child_branch is not None)
                     or (child_branch is not None and other_child_branch is None)):
                 return True
-            if child_branch.different_shape_to(other_child_branch):
+            # if child_branch is None then other_child_branch must be
+            # None too, meaning that they are both run instructions,
+            # we would compare their nest locations (commands), but
+            # that has already been done, so just guard
+            if (child_branch is not None
+                    and child_branch.different_shape_to(other_child_branch)):
                 return True
         return False
 
+    def _list_child_names(self):
+        child_names = []
+        for instruction in self.child_branches:
+            child_branch = instruction.recipe_branch
+            if child_branch is None:
+                continue
+            child_names += child_branch.list_branch_names()
+        return child_names
+
+    def list_branch_names(self):
+        """List all of the branch names under this one.
+
+        :return: a list of the branch names.
+        :rtype: list(str)
+        """
+        return [self.name] + self._list_child_names()
+
 
 class BaseRecipeBranch(RecipeBranch):
     """The RecipeBranch that is at the root of a recipe."""
@@ -545,6 +701,9 @@ class BaseRecipeBranch(RecipeBranch):
         if TIME_VAR in self.deb_version:
             self.deb_version = self.deb_version.replace(TIME_VAR,
                     time.strftime("%Y%m%d%H%M"))
+        if DATE_VAR in self.deb_version:
+            self.deb_version = self.deb_version.replace(DATE_VAR,
+                    time.strftime("%Y%m%d"))
 
     def substitute_debupstream(self, version):
         """Substitute {debupstream} in to deb_version if needed.
@@ -557,6 +716,43 @@ class BaseRecipeBranch(RecipeBranch):
             self.deb_version = self.deb_version.replace(DEBUPSTREAM_VAR,
                     version.upstream_version)
 
+    def _add_child_branches_to_manifest(self, child_branches, indent_level):
+        manifest = ""
+        for instruction in child_branches:
+            manifest += "%s%s\n" % (
+                "  " * indent_level, instruction.as_text())
+            if instruction.can_have_children:
+                manifest += self._add_child_branches_to_manifest(
+                    instruction.recipe_branch.child_branches,
+                    indent_level+1)
+        return manifest
+
+
+    def __str__(self):
+        return self.get_recipe_text(validate=True)
+
+    def list_branch_names(self):
+        return self._list_child_names()
+
+    def get_recipe_text(self, validate=False):
+        manifest = "# bzr-builder format %s deb-version " % str(self.format)
+        # TODO: should we store the expanded version that was used?
+        manifest += "%s\n" % (self.deb_version,)
+        if self.revid is not None:
+            manifest += "%s revid:%s\n" % (self.url, self.revid)
+        elif self.revspec is not None:
+            manifest += "%s %s\n" % (self.url, self.revspec)
+        else:
+            manifest += "%s\n" % (self.url,)
+        manifest += self._add_child_branches_to_manifest(self.child_branches,
+                0)
+        if validate:
+            # Sanity check.
+            # TODO: write a function that compares the result of this parse with
+            # the branch that we built it from.
+            RecipeParser(manifest).parse()
+        return manifest
+
 
 class RecipeParseError(errors.BzrError):
     _fmt = "Error parsing %(filename)s:%(line)s:%(char)s: %(problem)s."
@@ -566,6 +762,13 @@ class RecipeParseError(errors.BzrError):
                 problem=problem)
 
 
+class ForbiddenInstructionError(RecipeParseError):
+
+    def __init__(self, filename, line, char, problem, instruction_name=None):
+        RecipeParseError.__init__(self, filename, line, char, problem)
+        self.instruction_name = instruction_name
+
+
 class RecipeParser(object):
     """Parse a recipe.
 
@@ -576,7 +779,7 @@ class RecipeParser(object):
     eol_char = "\n"
     digit_chars = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")
 
-    NEWEST_VERSION = 0.2
+    NEWEST_VERSION = 0.3
 
     def __init__(self, f, filename=None):
         """Create a RecipeParser.
@@ -594,9 +797,12 @@ class RecipeParser(object):
         if filename is None:
             self.filename = "recipe"
 
-    def parse(self):
+    def parse(self, permitted_instructions=None):
         """Parse the recipe.
 
+        :param permitted_instructions: a list of instructions that you
+            want to allow. Defaults to None allowing them all.
+        :type permitted_instructions: list(str) or None
         :return: a RecipeBranch representing the recipe.
         """
         self.lines = self.text.split("\n")
@@ -604,12 +810,21 @@ class RecipeParser(object):
         self.line_index = 0
         self.current_line = self.lines[self.line_index]
         self.current_indent_level = 0
+        self.seen_nicks = set()
+        self.seen_paths = {".": 1}
         (version, deb_version) = self.parse_header()
         self.version = version
         last_instruction = None
         active_branches = []
         last_branch = None
         while self.line_index < len(self.lines):
+            if self.is_blankline():
+                self.new_line()
+                continue
+            comment = self.parse_comment_line()
+            if comment is not None:
+                self.new_line()
+                continue
             old_indent_level = self.parse_indent()
             if old_indent_level is not None:
                 if (old_indent_level < self.current_indent_level
@@ -621,10 +836,6 @@ class RecipeParser(object):
                 else:
                     unindent = self.current_indent_level - old_indent_level
                     active_branches = active_branches[:unindent]
-            comment = self.parse_comment_line()
-            if comment is not None:
-                self.new_line()
-                continue
             if last_instruction is None:
                 url = self.take_to_whitespace("branch to start from")
                 revspec = self.parse_optional_revspec()
@@ -634,7 +845,8 @@ class RecipeParser(object):
                 active_branches = [last_branch]
                 last_instruction = ""
             else:
-                instruction = self.parse_instruction()
+                instruction = self.parse_instruction(
+                    permitted_instructions=permitted_instructions)
                 if instruction == RUN_INSTRUCTION:
                     self.parse_whitespace("the command")
                     command = self.take_to_newline().strip()
@@ -645,13 +857,25 @@ class RecipeParser(object):
                     url = self.parse_branch_url()
                     if instruction == NEST_INSTRUCTION:
                         location = self.parse_branch_location()
-                    revspec = self.parse_optional_revspec()
+                    if instruction == NEST_PART_INSTRUCTION:
+                        path = self.parse_subpath()
+                        target_subdir = self.parse_optional_path()
+                        if target_subdir == '':
+                            target_subdir = None
+                            revspec = None
+                        else:
+                            revspec = self.parse_optional_revspec()
+                    else:
+                        revspec = self.parse_optional_revspec()
                     self.new_line()
                     last_branch = RecipeBranch(branch_id, url, revspec=revspec)
                     if instruction == NEST_INSTRUCTION:
                         active_branches[-1].nest_branch(location, last_branch)
-                    else:
+                    elif instruction == MERGE_INSTRUCTION:
                         active_branches[-1].merge_branch(last_branch)
+                    elif instruction == NEST_PART_INSTRUCTION:
+                        active_branches[-1].nest_part_branch(
+                            last_branch, path, target_subdir)
                 last_instruction = instruction
         if len(active_branches) == 0:
             self.throw_parse_error("Empty recipe")
@@ -671,18 +895,28 @@ class RecipeParser(object):
         self.new_line()
         return version, deb_version
 
-    def parse_instruction(self):
+    def parse_instruction(self, permitted_instructions=None):
         if self.version < 0.2:
             options = (MERGE_INSTRUCTION, NEST_INSTRUCTION)
             options_str = "'%s' or '%s'" % options
-        else:
+        elif self.version < 0.3:
             options = (MERGE_INSTRUCTION, NEST_INSTRUCTION, RUN_INSTRUCTION)
             options_str = "'%s', '%s' or '%s'" % options
+        else:
+            options = (MERGE_INSTRUCTION, NEST_INSTRUCTION,
+                NEST_PART_INSTRUCTION, RUN_INSTRUCTION)
+            options_str = "'%s', '%s', '%s' or '%s'" % options
         instruction = self.peek_to_whitespace()
         if instruction is None:
             self.throw_parse_error("End of line while looking for %s"
                     % options_str)
         if instruction in options:
+            if permitted_instructions is not None:
+                if instruction not in permitted_instructions:
+                    self.throw_parse_error("The '%s' instruction is "
+                            "forbidden" % instruction,
+                            cls=ForbiddenInstructionError,
+                            instruction_name=instruction)
             self.take_chars(len(instruction))
             return instruction
         self.throw_parse_error("Expecting %s, got '%s'"
@@ -690,7 +924,15 @@ class RecipeParser(object):
 
     def parse_branch_id(self):
         self.parse_whitespace("the branch id")
-        branch_id = self.take_to_whitespace("the branch id")
+        branch_id = self.peek_to_whitespace()
+        if branch_id is None:
+            self.throw_parse_error("End of line while looking for the "
+                    "branch id")
+        if branch_id in self.seen_nicks:
+            self.throw_parse_error("'%s' was already used to identify "
+                    "a branch." % branch_id)
+        self.take_chars(len(branch_id))
+        self.seen_nicks.add(branch_id)
         return branch_id
 
     def parse_branch_url(self):
@@ -701,9 +943,35 @@ class RecipeParser(object):
     def parse_branch_location(self):
         # FIXME: Needs a better term
         self.parse_whitespace("the location to nest")
-        location = self.take_to_whitespace("the location to nest")
+        location = self.peek_to_whitespace()
+        if location is None:
+            self.throw_parse_error("End of line while looking for the "
+                    "location to nest")
+        norm_location = os.path.normpath(location)
+        if norm_location in self.seen_paths:
+            self.throw_parse_error("The path '%s' is a duplicate of "
+                    "the one used on line %d." % (location,
+                        self.seen_paths[norm_location]))
+        if os.path.isabs(norm_location):
+            self.throw_parse_error("Absolute paths are not allowed: %s"
+                    % location)
+        if norm_location.startswith(".."):
+            self.throw_parse_error("Paths outside the current directory "
+                    "are not allowed: %s" % location)
+        self.take_chars(len(location))
+        self.seen_paths[norm_location] = self.line_index + 1
         return location
 
+    def parse_subpath(self):
+        self.parse_whitespace("the subpath to merge")
+        location = self.take_to_whitespace("the subpath to merge")
+        return location
+
+    def parse_revspec(self):
+        self.parse_whitespace("the revspec")
+        revspec = self.take_to_whitespace("the revspec")
+        return revspec
+
     def parse_optional_revspec(self):
         self.parse_whitespace(None, require=False)
         revspec = self.peek_to_whitespace()
@@ -711,9 +979,18 @@ class RecipeParser(object):
             self.take_chars(len(revspec))
         return revspec
 
-    def throw_parse_error(self, problem):
-        raise RecipeParseError(self.filename, self.line_index + 1,
-                self.index + 1, problem)
+    def parse_optional_path(self):
+        self.parse_whitespace(None, require=False)
+        path = self.peek_to_whitespace()
+        if path is not None:
+            self.take_chars(len(path))
+        return path
+
+    def throw_parse_error(self, problem, cls=None, **kwargs):
+        if cls is None:
+            cls = RecipeParseError
+        raise cls(self.filename, self.line_index + 1,
+                self.index + 1, problem, **kwargs)
 
     def throw_expecting_error(self, expected, actual):
         self.throw_parse_error("Expecting '%s', got '%s'"
@@ -736,6 +1013,12 @@ class RecipeParser(object):
         else:
             self.current_line = self.lines[self.line_index]
 
+    def is_blankline(self):
+        whitespace = self.peek_whitespace()
+        if whitespace is None:
+            return True
+        return self.peek_char(skip=len(whitespace)) is None
+
     def take_char(self):
         if self.index >= len(self.current_line):
             return None
@@ -805,6 +1088,18 @@ class RecipeParser(object):
             actual = self.peek_char()
         return ret
 
+    def peek_whitespace(self):
+        ret = ""
+        char = self.peek_char()
+        if char is None:
+            return char
+        count = 0
+        while char is not None and char in self.whitespace_chars:
+            ret += char
+            count += 1
+            char = self.peek_char(skip=count)
+        return ret
+
     def parse_word(self, expected, require_whitespace=True):
         self.parse_whitespace("'%s'" % expected, require=require_whitespace)
         length = len(expected)
@@ -881,10 +1176,14 @@ class RecipeParser(object):
         return text
 
     def parse_comment_line(self):
-        if self.peek_char() is None:
+        whitespace = self.peek_whitespace()
+        if whitespace is None:
             return ""
-        if self.peek_char() != "#":
+        if self.peek_char(skip=len(whitespace)) is None:
+            return ""
+        if self.peek_char(skip=len(whitespace)) != "#":
             return None
+        self.parse_whitespace(None, require=False)
         comment = self.current_line[self.index:]
         self.index += len(comment)
         return comment
