@@ -20,9 +20,10 @@ def test_suite():
     loader = TestUtil.TestLoader()
     suite = TestSuite()
     testmod_names = [
-            'test_blackbox',
-            'test_recipe',
+            'blackbox',
+            'ppa',
+            'recipe',
             ]
-    suite.addTest(loader.loadTestsFromModuleNames(["%s.%s" % (__name__, i)
+    suite.addTest(loader.loadTestsFromModuleNames(["%s.test_%s" % (__name__, i)
                                             for i in testmod_names]))
     return suite
