@@ -1,5 +1,5 @@
-# bzr-builter: a bzr plugin to constuct trees based on recipes
-# Copyright 2009 Canonical Ltd.
+# bzr-builder: a bzr plugin to constuct trees based on recipes
+# Copyright 2009-2010 Canonical Ltd.
 
 # This program is free software: you can redistribute it and/or modify it 
 # under the terms of the GNU General Public License version 3, as published 
@@ -55,7 +55,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         self.build_tree(["source/a"])
         source.add(["a"])
         revid = source.commit("one")
-        self.run_bzr("build recipe working")
+        self.run_bzr("build -q recipe working")
         self.failUnlessExists("working/a")
         tree = workingtree.WorkingTree.open("working")
         self.assertEqual(revid, tree.last_revision())
@@ -71,7 +71,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         self.build_tree(["source/a"])
         source.add(["a"])
         revid = source.commit("one")
-        self.run_bzr("build recipe working --manifest manifest")
+        self.run_bzr("build -q recipe working --manifest manifest")
         self.failUnlessExists("working/a")
         self.failUnlessExists("manifest")
         self.check_file_contents("manifest", "# bzr-builder format 0.1 "
@@ -111,7 +111,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
                     "deb-version 1\nsource 1\n")])
         self.build_tree_contents([("old-manifest", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource revid:foo\n")])
-        out, err = self.run_bzr("build recipe working --manifest manifest "
+        out, err = self.run_bzr("build -q recipe working --manifest manifest "
                 "--if-changed-from old-manifest")
         self.failUnlessExists("working/a")
         self.failUnlessExists("manifest")
@@ -130,7 +130,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         revid = source.commit("one")
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe working "
+        out, err = self.run_bzr("dailydeb -q test.recipe working "
                 "--manifest manifest --package foo")
         self.failIfExists("working/a")
         package_root = "working/foo-1/"
@@ -167,7 +167,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         source.commit("one")
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe "
+        out, err = self.run_bzr("dailydeb -q test.recipe "
                 "--manifest manifest --package foo")
 
     def test_cmd_dailydeb_if_changed_from_non_existant(self):
@@ -184,7 +184,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         source.commit("one")
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe "
+        out, err = self.run_bzr("dailydeb -q test.recipe "
                 "--manifest manifest --package foo --if-changed-from bar")
 
     def make_simple_package(self):
@@ -207,7 +207,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         self.make_simple_package()
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe "
+        out, err = self.run_bzr("dailydeb -q test.recipe "
                 "--manifest manifest --no-build working")
         new_cl_contents = ("package (1) unstable; urgency=low\n\n"
                 "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
@@ -225,7 +225,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         self.make_simple_package()
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version 1\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe "
+        out, err = self.run_bzr("dailydeb -q test.recipe "
                 "--manifest manifest --if-changed-from bar working")
         new_cl_contents = ("package (1) unstable; urgency=low\n\n"
                 "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
@@ -240,7 +240,7 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         self.make_simple_package()
         self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
                     "deb-version {debupstream}-2\nsource 1\n")])
-        out, err = self.run_bzr("dailydeb test.recipe working")
+        out, err = self.run_bzr("dailydeb -q test.recipe working")
         new_cl_contents = ("package (0.1-2) unstable; urgency=low\n\n"
                 "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
         f = open("working/test-{debupstream}-2/debian/changelog")
@@ -249,3 +249,38 @@ class BlackboxBuilderTests(TestCaseWithTransport):
         finally:
             f.close()
         self.assertStartsWith(actual_cl_contents, new_cl_contents)
+
+    def test_cmd_dailydeb_with_append_version(self):
+        self.make_simple_package()
+        self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
+                    "deb-version 1\nsource 1\n")])
+        out, err = self.run_bzr("dailydeb -q test.recipe working "
+                "--append-version ~ppa1")
+        new_cl_contents = ("package (1~ppa1) unstable; urgency=low\n\n"
+                "  * Auto build.\n\n -- M. Maintainer <maint@maint.org>  ")
+        f = open("working/test-1/debian/changelog")
+        try:
+            actual_cl_contents = f.read()
+        finally:
+            f.close()
+        self.assertStartsWith(actual_cl_contents, new_cl_contents)
+
+    def test_cmd_dailydeb_with_invalid_version(self):
+        source = self.make_branch_and_tree("source")
+        self.build_tree(["source/a"])
+        source.add(["a"])
+        revid = source.commit("one")
+        self.build_tree_contents([("test.recipe", "# bzr-builder format 0.1 "
+                    "deb-version $\nsource 1\n")])
+        err = self.run_bzr("dailydeb -q test.recipe working --package foo",
+                retcode=3)[1]
+        self.assertContainsRe(err, "bzr: ERROR: Invalid deb-version: \$: ")
+
+    def test_cmd_dailydeb_with_safe(self):
+        self.make_simple_package()
+        self.build_tree_contents([("test.recipe", "# bzr-builder format 0.3 "
+                    "deb-version 1\nsource 1\nrun something bad")])
+        out, err = self.run_bzr("dailydeb -q test.recipe working --safe",
+            retcode=3)
+        self.assertContainsRe(err, "The 'run' instruction is forbidden.$")
+
