@@ -27,26 +27,28 @@ from bzrlib.tests import (
         )
 from bzrlib.plugins.builder.recipe import (
         BaseRecipeBranch,
-        build_manifest,
         build_tree,
         ensure_basedir,
+        ForbiddenInstructionError,
         pull_or_branch,
         RecipeParser,
         RecipeBranch,
         RecipeParseError,
         resolve_revisions,
+        SAFE_INSTRUCTIONS,
         )
 
 
 class RecipeParserTests(TestCaseInTempDir):
 
     deb_version = "0.1-{revno}"
-    basic_header = ("# bzr-builder format 0.2 deb-version "
+    basic_header = ("# bzr-builder format 0.3 deb-version "
             + deb_version +"\n")
-    basic_header_and_branch = basic_header + "http://foo.org/\n"
+    basic_branch = "http://foo.org/"
+    basic_header_and_branch = basic_header + basic_branch + "\n"
 
-    def get_recipe(self, recipe_text):
-        return RecipeParser(recipe_text).parse()
+    def get_recipe(self, recipe_text, **kwargs):
+        return RecipeParser(recipe_text).parse(**kwargs)
 
     def assertParseError(self, line, char, problem, callable, *args,
             **kwargs):
@@ -55,6 +57,7 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertEqual(line, exc.line)
         self.assertEqual(char, exc.char)
         self.assertEqual("recipe", exc.filename)
+        return exc
 
     def check_recipe_branch(self, branch, name, url, revspec=None,
             num_child_branches=0, revid=None):
@@ -130,8 +133,8 @@ class RecipeParserTests(TestCaseInTempDir):
                 self.basic_header + "http://foo.org/ 2 foo")
 
     def tests_rejects_unknown_instruction(self):
-        self.assertParseError(3, 1, "Expecting 'merge', 'nest' or 'run', "
-                "got 'cat'", self.get_recipe,
+        self.assertParseError(3, 1, "Expecting 'merge', 'nest', 'nest-part' "
+                "or 'run', got 'cat'", self.get_recipe,
                 self.basic_header + "http://foo.org/\n" + "cat")
 
     def test_rejects_merge_no_name(self):
@@ -149,6 +152,21 @@ class RecipeParserTests(TestCaseInTempDir):
                 "got 'bar'", self.get_recipe,
                 self.basic_header_and_branch + "merge foo url 2 bar")
 
+    def test_rejects_nest_part_no_name(self):
+        self.assertParseError(3, 11, "End of line while looking for "
+                "the branch id", self.get_recipe,
+                self.basic_header_and_branch + "nest-part ")
+
+    def test_rejects_nest_part_no_url(self):
+        self.assertParseError(3, 15, "End of line while looking for "
+                "the branch url", self.get_recipe,
+                self.basic_header_and_branch + "nest-part foo ")
+
+    def test_rejects_nest_part_no_subpath(self):
+        self.assertParseError(3, 22, "End of line while looking for "
+                "the subpath to merge", self.get_recipe,
+                self.basic_header_and_branch + "nest-part foo url:// ")
+
     def test_rejects_nest_no_name(self):
         self.assertParseError(3, 6, "End of line while looking for "
                 "the branch id", self.get_recipe,
@@ -202,6 +220,12 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertParseError(3, 1, "Empty recipe", self.get_recipe,
                 self.basic_header)
 
+    def test_rejects_non_unique_ids(self):
+        self.assertParseError(4, 7, "'foo' was already used to identify "
+                "a branch.", self.get_recipe,
+                self.basic_header_and_branch + "merge foo url\n"
+                + "merge foo other-url\n")
+
     def test_builds_simplest_recipe(self):
         base_branch = self.get_recipe(self.basic_header_and_branch)
         self.check_base_recipe_branch(base_branch, "http://foo.org/")
@@ -220,6 +244,42 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertEqual(None, location)
         self.check_recipe_branch(child_branch, "bar", "http://bar.org")
 
+    def test_builds_recipe_with_nest_part(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest-part bar http://bar.org some/path")
+        self.check_base_recipe_branch(base_branch, "http://foo.org/",
+                num_child_branches=1)
+        instruction = base_branch.child_branches[0]
+        self.assertEqual(None, instruction.nest_path)
+        self.check_recipe_branch(
+            instruction.recipe_branch, "bar", "http://bar.org")
+        self.assertEqual("some/path", instruction.subpath)
+        self.assertEqual(None, instruction.target_subdir)
+
+    def test_builds_recipe_with_nest_part_subdir(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest-part bar http://bar.org some/path target-subdir")
+        self.check_base_recipe_branch(base_branch, "http://foo.org/",
+                num_child_branches=1)
+        instruction = base_branch.child_branches[0]
+        self.assertEqual(None, instruction.nest_path)
+        self.check_recipe_branch(
+            instruction.recipe_branch, "bar", "http://bar.org")
+        self.assertEqual("some/path", instruction.subpath)
+        self.assertEqual("target-subdir", instruction.target_subdir)
+
+    def test_builds_recipe_with_nest_part_subdir_and_revspec(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest-part bar http://bar.org some/path target-subdir 1234")
+        self.check_base_recipe_branch(base_branch, "http://foo.org/",
+                num_child_branches=1)
+        instruction = base_branch.child_branches[0]
+        self.assertEqual(None, instruction.nest_path)
+        self.check_recipe_branch(
+            instruction.recipe_branch, "bar", "http://bar.org", "1234")
+        self.assertEqual("some/path", instruction.subpath)
+        self.assertEqual("target-subdir", instruction.target_subdir)
+
     def test_builds_recipe_with_nest(self):
         base_branch = self.get_recipe(self.basic_header_and_branch
                 + "nest bar http://bar.org baz")
@@ -306,6 +366,60 @@ class RecipeParserTests(TestCaseInTempDir):
         self.assertEqual(None, child_branch)
         self.assertEqual("touch test", command)
 
+    def test_accepts_blank_line_during_nest(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest foo http://bar.org bar\n  merge baz baz.org\n\n"
+                "  merge zap zap.org\n")
+        self.check_base_recipe_branch(base_branch, self.basic_branch,
+                num_child_branches=1)
+        nested_branch, location = base_branch.child_branches[0].as_tuple()
+        self.assertEqual("bar", location)
+        self.check_recipe_branch(nested_branch, "foo", "http://bar.org",
+                num_child_branches=2)
+        child_branch, location = nested_branch.child_branches[0].as_tuple()
+        self.assertEqual(None, location)
+        self.check_recipe_branch(child_branch, "baz", "baz.org")
+        child_branch, location = nested_branch.child_branches[1].as_tuple()
+        self.assertEqual(None, location)
+        self.check_recipe_branch(child_branch, "zap", "zap.org")
+
+    def test_accepts_blank_line_at_start_of_nest(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest foo http://bar.org bar\n\n  merge baz baz.org\n")
+        self.check_base_recipe_branch(base_branch, self.basic_branch,
+                num_child_branches=1)
+        nested_branch, location = base_branch.child_branches[0].as_tuple()
+        self.assertEqual("bar", location)
+        self.check_recipe_branch(nested_branch, "foo", "http://bar.org",
+                num_child_branches=1)
+        child_branch, location = nested_branch.child_branches[0].as_tuple()
+        self.assertEqual(None, location)
+        self.check_recipe_branch(child_branch, "baz", "baz.org")
+
+    def test_accepts_blank_line_as_only_thing_in_nest(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest foo http://bar.org bar\n\nmerge baz baz.org\n")
+        self.check_base_recipe_branch(base_branch, self.basic_branch,
+                num_child_branches=2)
+        nested_branch, location = base_branch.child_branches[0].as_tuple()
+        self.assertEqual("bar", location)
+        self.check_recipe_branch(nested_branch, "foo", "http://bar.org")
+        child_branch, location = base_branch.child_branches[1].as_tuple()
+        self.assertEqual(None, location)
+        self.check_recipe_branch(child_branch, "baz", "baz.org")
+
+    def test_accepts_comment_line_with_any_number_of_spaces(self):
+        base_branch = self.get_recipe(self.basic_header_and_branch
+                + "nest foo http://bar.org bar\n   #foo\nmerge baz baz.org\n")
+        self.check_base_recipe_branch(base_branch, self.basic_branch,
+                num_child_branches=2)
+        nested_branch, location = base_branch.child_branches[0].as_tuple()
+        self.assertEqual("bar", location)
+        self.check_recipe_branch(nested_branch, "foo", "http://bar.org")
+        child_branch, location = base_branch.child_branches[1].as_tuple()
+        self.assertEqual(None, location)
+        self.check_recipe_branch(child_branch, "baz", "baz.org")
+
     def test_old_format_rejects_run(self):
         header = ("# bzr-builder format 0.1 deb-version "
                 + self.deb_version +"\n")
@@ -313,6 +427,52 @@ class RecipeParserTests(TestCaseInTempDir):
                 , self.get_recipe, header + "http://foo.org/\n"
                 + "run touch test \n")
 
+    def test_old_format_rejects_nest_part(self):
+        header = ("# bzr-builder format 0.2 deb-version "
+                + self.deb_version +"\n")
+        self.assertParseError(3, 1, "Expecting 'merge', 'nest' or 'run', "
+                "got 'nest-part'" , self.get_recipe,
+                header + "http://foo.org/\nnest-part packaging foo test \n")
+
+    def test_error_on_forbidden_instructions(self):
+        exc = self.assertParseError(3, 1, "The 'run' instruction is "
+                "forbidden", self.get_recipe, self.basic_header_and_branch
+                + "run touch test\n",
+                permitted_instructions=[SAFE_INSTRUCTIONS])
+        self.assertTrue(isinstance(exc, ForbiddenInstructionError))
+        self.assertEqual("run", exc.instruction_name)
+
+    def test_error_on_duplicate_path(self):
+        exc = self.assertParseError(3, 15, "The path '.' is a duplicate "
+                "of the one used on line 1.", self.get_recipe,
+                self.basic_header_and_branch + "nest nest url .\n")
+
+    def test_error_on_duplicate_path_with_another_nest(self):
+        exc = self.assertParseError(4, 16, "The path 'foo' is a duplicate "
+                "of the one used on line 3.", self.get_recipe,
+                self.basic_header_and_branch + "nest nest url foo\n"
+                + "nest nest2 url foo\n")
+
+    def test_duplicate_path_check_uses_normpath(self):
+        exc = self.assertParseError(3, 15, "The path 'foo/..' is a duplicate "
+                "of the one used on line 1.", self.get_recipe,
+                self.basic_header_and_branch + "nest nest url foo/..\n")
+
+    def test_error_absolute_path(self):
+        exc = self.assertParseError(3, 15, "Absolute paths are not allowed: "
+                "/etc/passwd", self.get_recipe, self.basic_header_and_branch
+                + "nest nest url /etc/passwd\n")
+
+    def test_error_simple_parent_dir(self):
+        exc = self.assertParseError(3, 15, "Paths outside the current "
+                "directory are not allowed: ../foo", self.get_recipe,
+                self.basic_header_and_branch + "nest nest url ../foo\n")
+
+    def test_error_complex_parent_dir(self):
+        exc = self.assertParseError(3, 15, "Paths outside the current "
+                "directory are not allowed: ./foo/../..", self.get_recipe,
+                self.basic_header_and_branch + "nest nest url ./foo/../..\n")
+
 
 class BuildTreeTests(TestCaseWithTransport):
 
@@ -360,15 +520,17 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
 
-    def test_build_tree_nested(self):
-        source1 = self.make_branch_and_tree("source1")
-        self.build_tree(["source1/a"])
+    def make_source_branch(self, relpath):
+        """Make a branch with one file and one commit."""
+        source1 = self.make_branch_and_tree(relpath)
+        self.build_tree([relpath + "/a"])
         source1.add(["a"])
-        source1_rev_id = source1.commit("one")
-        source2 = self.make_branch_and_tree("source2")
-        self.build_tree(["source2/a"])
-        source2.add(["a"])
-        source2_rev_id = source2.commit("one")
+        source1.commit("one")
+        return source1
+
+    def test_build_tree_nested(self):
+        source1_rev_id = self.make_source_branch("source1").last_revision()
+        source2_rev_id = self.make_source_branch("source2").last_revision()
         base_branch = BaseRecipeBranch("source1", "1", 0.2)
         nested_branch = RecipeBranch("nested", "source2")
         base_branch.nest_branch("sub", nested_branch)
@@ -382,10 +544,8 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(source2_rev_id, nested_branch.revid)
 
     def test_build_tree_merged(self):
-        source1 = self.make_branch_and_tree("source1")
-        self.build_tree(["source1/a"])
-        source1.add(["a"])
-        source1_rev_id = source1.commit("one")
+        source1 = self.make_source_branch("source1")
+        source1_rev_id = source1.last_revision()
         source2 = source1.bzrdir.sprout("source2").open_workingtree()
         self.build_tree_contents([("source2/a", "other change")])
         source2_rev_id = source2.commit("one")
@@ -403,11 +563,60 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(source1_rev_id, base_branch.revid)
         self.assertEqual(source2_rev_id, merged_branch.revid)
 
-    def test_build_tree_merge_twice(self):
+    def test_build_tree_nest_part(self):
+        """A recipe can specify a merge of just part of an unrelated tree."""
+        source1 = self.make_source_branch("source1")
+        source2 = self.make_source_branch("source2")
+        source1.lock_read()
+        self.addCleanup(source1.unlock)
+        source1_rev_id = source1.last_revision()
+        # Add 'b' to source2.
+        self.build_tree_contents([
+            ("source2/b", "new file"), ("source2/not-b", "other file")])
+        source2.add(["b", "not-b"])
+        source2_rev_id = source2.commit("two")
+        base_branch = BaseRecipeBranch("source1", "1", 0.2)
+        merged_branch = RecipeBranch("merged", "source2")
+        # Merge just 'b' from source2; 'a' is untouched.
+        base_branch.nest_part_branch(merged_branch, "b")
+        build_tree(base_branch, "target")
+        file_id = source1.path2id("a")
+        self.check_file_contents("target/a", source1.get_file_text(file_id))
+        self.check_file_contents("target/b", "new file")
+        self.assertNotInWorkingTree("not-b", "target")
+        self.assertEqual(source1_rev_id, base_branch.revid)
+        self.assertEqual(source2_rev_id, merged_branch.revid)
+
+    def test_build_tree_nest_part_explicit_target(self):
+        """A recipe can specify a merge of just part of an unrelated tree into
+        a specific subdirectory of the target tree.
+        """
         source1 = self.make_branch_and_tree("source1")
-        self.build_tree(["source1/a"])
-        source1.add(["a"])
-        source1_rev_id = source1.commit("one")
+        self.build_tree(["source1/dir/"])
+        source1.add(["dir"])
+        source1.commit("one")
+        source2 = self.make_source_branch("source2")
+        source1.lock_read()
+        self.addCleanup(source1.unlock)
+        source1_rev_id = source1.last_revision()
+        # Add 'b' to source2.
+        self.build_tree_contents([
+            ("source2/b", "new file"), ("source2/not-b", "other file")])
+        source2.add(["b", "not-b"])
+        source2_rev_id = source2.commit("two")
+        base_branch = BaseRecipeBranch("source1", "1", 0.2)
+        merged_branch = RecipeBranch("merged", "source2")
+        # Merge just 'b' from source2; 'a' is untouched.
+        base_branch.nest_part_branch(merged_branch, "b", "dir/b")
+        build_tree(base_branch, "target")
+        self.check_file_contents("target/dir/b", "new file")
+        self.assertNotInWorkingTree("dir/not-b", "target")
+        self.assertEqual(source1_rev_id, base_branch.revid)
+        self.assertEqual(source2_rev_id, merged_branch.revid)
+
+    def test_build_tree_merge_twice(self):
+        source1 = self.make_source_branch("source1")
+        source1_rev_id = source1.last_revision()
         source2 = source1.bzrdir.sprout("source2").open_workingtree()
         self.build_tree_contents([("source2/a", "other change")])
         source2_rev_id = source2.commit("one")
@@ -436,10 +645,7 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(source3_rev_id, merged_branch2.revid)
 
     def test_build_tree_merged_with_conflicts(self):
-        source1 = self.make_branch_and_tree("source1")
-        self.build_tree(["source1/a"])
-        source1.add(["a"])
-        source1_rev_id = source1.commit("one")
+        source1 = self.make_source_branch("source1")
         source2 = source1.bzrdir.sprout("source2").open_workingtree()
         self.build_tree_contents([("source2/a", "other change\n")])
         source2_rev_id = source2.commit("one")
@@ -465,10 +671,8 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(source2_rev_id, merged_branch.revid)
 
     def test_build_tree_with_revspecs(self):
-        source1 = self.make_branch_and_tree("source1")
-        self.build_tree(["source1/a"])
-        source1.add(["a"])
-        source1_rev_id = source1.commit("one")
+        source1 = self.make_source_branch("source1")
+        source1_rev_id = source1.last_revision()
         source2 = source1.bzrdir.sprout("source2").open_workingtree()
         self.build_tree_contents([("source2/a", "other change\n")])
         source2_rev_id = source2.commit("one")
@@ -622,6 +826,20 @@ class BuildTreeTests(TestCaseWithTransport):
         self.assertEqual(revid, tree.last_revision())
         self.assertEqual(revid, base_branch.revid)
 
+    def test_error_on_merge_revspec(self):
+        # See bug 416950
+        source = self.make_branch_and_tree("source")
+        revid = source.commit("one")
+        base_branch = BaseRecipeBranch("source", "1", 0.2)
+        merged_branch = RecipeBranch("merged", "source", revspec="debian")
+        base_branch.merge_branch(merged_branch)
+        e = self.assertRaises(errors.InvalidRevisionSpec,
+                build_tree, base_branch, "target")
+        self.assertTrue(str(e).startswith("Requested revision: 'debian' "
+                    "does not exist in branch: "))
+        self.assertTrue(str(e).endswith(". Did you not mean to specify a "
+                    "revspec at the end of the merge line?"))
+
 
 class ResolveRevisionsTests(TestCaseWithTransport):
 
@@ -713,6 +931,17 @@ class ResolveRevisionsTests(TestCaseWithTransport):
                     if_changed_from=branch2))
         self.assertEqual("source", branch1.url)
 
+    def test_unchanged_command(self):
+        source =self.make_branch_and_tree("source")
+        source.commit("one")
+        branch1 = BaseRecipeBranch("source", "{revno}", 0.2)
+        branch2 = BaseRecipeBranch("source", "{revno}", 0.2)
+        branch1.run_command("touch test1")
+        branch2.run_command("touch test1")
+        self.assertEqual(False, resolve_revisions(branch1,
+                    if_changed_from=branch2))
+        self.assertEqual("source", branch1.url)
+
     def test_substitute(self):
         source =self.make_branch_and_tree("source")
         revid1 = source.commit("one")
@@ -737,13 +966,20 @@ class ResolveRevisionsTests(TestCaseWithTransport):
         resolve_revisions(branch1)
         self.assertEqual("{debupstream}-2", branch1.deb_version)
 
+    def test_subsitute_not_fully_expanded(self):
+        source =self.make_branch_and_tree("source")
+        source.commit("one")
+        source.commit("two")
+        branch1 = BaseRecipeBranch("source", "{revno:packaging}", 0.2)
+        self.assertRaises(errors.BzrCommandError, resolve_revisions, branch1)
 
-class BuildManifestTests(TestCaseInTempDir):
+
+class StringifyTests(TestCaseInTempDir):
 
     def test_simple_manifest(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.1)
         base_branch.revid = "base_revid"
-        manifest = build_manifest(base_branch)
+        manifest = str(base_branch)
         self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
                 "base_url revid:base_revid\n", manifest)
 
@@ -759,7 +995,7 @@ class BuildManifestTests(TestCaseInTempDir):
         merged_branch = RecipeBranch("merged", "merged_url")
         merged_branch.revid = "merged_revid"
         base_branch.merge_branch(merged_branch)
-        manifest = build_manifest(base_branch)
+        manifest = str(base_branch)
         self.assertEqual("# bzr-builder format 0.2 deb-version 2\n"
                 "base_url revid:base_revid\n"
                 "nest nested1 nested1_url nested revid:nested1_revid\n"
@@ -770,11 +1006,87 @@ class BuildManifestTests(TestCaseInTempDir):
         base_branch = BaseRecipeBranch("base_url", "1", 0.2)
         base_branch.revid = "base_revid"
         base_branch.run_command("touch test")
-        manifest = build_manifest(base_branch)
+        manifest = str(base_branch)
         self.assertEqual("# bzr-builder format 0.2 deb-version 1\n"
                 "base_url revid:base_revid\n"
                 "run touch test\n", manifest)
 
+    def test_recipe_with_no_revspec(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        manifest = str(base_branch)
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url\n", manifest)
+
+    def test_recipe_with_tag_revspec(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1,
+                revspec="tag:foo")
+        manifest = str(base_branch)
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url tag:foo\n", manifest)
+
+    def test_recipe_with_child(self):
+        base_branch = BaseRecipeBranch("base_url", "2", 0.2)
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_branch("nested", nested_branch1)
+        nested_branch2 = RecipeBranch("nested2", "nested2_url")
+        nested_branch1.nest_branch("nested2", nested_branch2)
+        merged_branch = RecipeBranch("merged", "merged_url")
+        base_branch.merge_branch(merged_branch)
+        manifest = str(base_branch)
+        self.assertEqual("# bzr-builder format 0.2 deb-version 2\n"
+                "base_url\n"
+                "nest nested1 nested1_url nested tag:foo\n"
+                "  nest nested2 nested2_url nested2\n"
+                "merge merged merged_url\n", manifest)
+
+    def test_get_recipe_text(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        manifest = base_branch.get_recipe_text()
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url revid:base_revid\n", manifest)
+
+    def test_get_recipe_doesnt_raise_on_invalid_recipe(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_branch(".", nested_branch1)
+        manifest = base_branch.get_recipe_text()
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url revid:base_revid\n"
+                "nest nested1 nested1_url . tag:foo\n", manifest)
+
+    def test_get_recipe_text_validate_True(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_branch(".", nested_branch1)
+        self.assertRaises(RecipeParseError, base_branch.get_recipe_text,
+                validate=True)
+
+    def test_str_validates(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_branch(".", nested_branch1)
+        self.assertRaises(RecipeParseError, str, base_branch)
+
+    def test_with_nest_part(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.1)
+        base_branch.revid = "base_revid"
+        nested_branch1 = RecipeBranch("nested1", "nested1_url",
+                revspec="tag:foo")
+        base_branch.nest_part_branch(nested_branch1, "foo", "bar")
+        manifest = base_branch.get_recipe_text()
+        self.assertEqual("# bzr-builder format 0.1 deb-version 1\n"
+                "base_url revid:base_revid\n"
+                "nest-part nested1 nested1_url foo bar tag:foo\n",
+                manifest)
+
 
 class RecipeBranchTests(TestCaseInTempDir):
 
@@ -817,6 +1129,14 @@ class RecipeBranchTests(TestCaseInTempDir):
         base_branch.substitute_time(time)
         self.assertEqual("1-197001010000", base_branch.deb_version)
 
+    def test_substitute_date(self):
+        time = datetime.datetime.utcfromtimestamp(1)
+        base_branch = BaseRecipeBranch("base_url", "1-{date}", 0.2)
+        base_branch.substitute_time(time)
+        self.assertEqual("1-19700101", base_branch.deb_version)
+        base_branch.substitute_time(time)
+        self.assertEqual("1-19700101", base_branch.deb_version)
+
     def test_substitute_revno(self):
         base_branch = BaseRecipeBranch("base_url", "1", 0.2)
         base_branch.substitute_revno(None, None)
@@ -838,3 +1158,17 @@ class RecipeBranchTests(TestCaseInTempDir):
         self.assertEqual("3", base_branch.deb_version)
         base_branch.substitute_revno("foo", lambda: "3")
         self.assertEqual("3", base_branch.deb_version)
+
+    def test_list_branch_names(self):
+        base_branch = BaseRecipeBranch("base_url", "1", 0.2)
+        base_branch.merge_branch(RecipeBranch("merged", "merged_url"))
+        nested_branch = RecipeBranch("nested", "nested_url")
+        nested_branch.merge_branch(
+            RecipeBranch("merged_into_nested", "another_url"))
+        base_branch.nest_branch("subdir", nested_branch)
+        base_branch.merge_branch(
+            RecipeBranch("another_nested", "yet_another_url"))
+        base_branch.run_command("a command")
+        self.assertEqual(
+            ["merged", "nested", "merged_into_nested", "another_nested"],
+            base_branch.list_branch_names())
